/*wch_tools
file converter tools for use with wch controller
features: 
 - input:   up to 3 intel hex files
 - output:  intel hex and raw binary
 - optional randomize unused bytes
 - optional TEA or XTEA encryption with adjustable key
 - optional use strong encryption
 - optional MSB first encryption (usefull for keil) 
 - optinal  limit output size
 - optional automatic flashing (not yet implemented)
*/

#include <stdio.h>  
#include <string.h> 
#include <stdlib.h>
#include <ctype.h>
#include <winsock.h>  // ltons()

#define MAX_STRING_SIZE 256
#define MAX_HEX_FILES   4

// globals
unsigned char FlashBuf[0xF400]; //max size for a ch559 
unsigned char InFname [MAX_STRING_SIZE] = "\0";
unsigned char OutFname[MAX_STRING_SIZE]   = "FIRMWARE.BIN\0";

unsigned char EncKeyLsb[MAX_STRING_SIZE]  = "TheQuickBrownFox\0";
unsigned char EncKeyMsb[MAX_STRING_SIZE]  = "QehTkciuworBxiFn\0";

int           WchCpu        = 0x59;
int           UserFlash     = 0xF400;
int           LimitSize     = 0xF400;
int           Encode        = 0;
int           XEncode       = 0;  
int           FillRandom    = 0;
int           AutoFlash     = 0;
int           MsbFirst      = 0;
int           StrongEnc     = 0;
   
// a short help screen
void DisplayHelp(void)
{
   printf("\noptions :\n");
   printf("   -i<INFILE>  : Intel Hex Input file[s]    mantory \n");
   printf("   -o<OUTFILE> : Binary outputfile          default: FIRMWARE.BIN\n");
   printf("   -c<CPU>     : WCH cpu typ (CPU id)       default: 0x59 for CH559\n");
   printf("   -l<LIMIT>   : Limit size of output       default: 0xF400 for a CH559 depending on CPU\n");
   printf("   -k<TEAKEY>  : TEA Encryption Key         default: \"TheQuickBrownFox\"\n");
   printf("   -e<+|->     : TEA Encryption     on|off  default: e- encrypt off -e+ implies -x-\n");
   printf("   -x<+|->     : XTEA Encryption    on|off  default: x- encrypt off -x+ implies -e-\n");
   printf("   -s<+|->     : Strong encryption  on|off  default: s- fast encryption\n");
   printf("   -m<+|->     : Msb first (Keil)   on|off  default: m- lsb first \n");
   printf("   -f<+|->     : Random fill        on|off  default: f- fill with 0xFF\n"); 
   printf("   -a<+|->     : Automatic Flashing on|off  default: a- Autoflash off\n");
   printf("\nexample:\n");
   printf("   wch_tools -iBlablub.hex,test.hex,main.hex -f+ -c0x52 -e+\n");
}

// reads a intel hex file and saves the data into the FlashBUf
int HexReader(unsigned char *filename)
{
   FILE           *file;
   unsigned char  rec_length, rec_type, rec_chksum, chksum;
   char           line[256],tmp[5];
   unsigned short rec_addr;
   int            i;
   int            warning = 1;
     
   if (filename)
   { 
      if (filename[0] == '\0') return -1; 
      file = fopen(filename,"r");
      if(file)
      {
         while(fgets(line,255,file))
         {
            if(line[0] != ':') 
            {
               printf("Error:  not a Intel Hex file\n"); 
               fclose(file);
               return -1;
            }
            memcpy(tmp,&line[1],2);
            tmp[2] = '\0';
            rec_length = (unsigned char)strtol(tmp,NULL,16);
            memcpy(tmp,&line[3],4);
            tmp[4] = '\0';
            rec_addr = (unsigned short)strtol(tmp,NULL,16);
            memcpy(tmp,&line[7],2);
            tmp[2] = '\0';
            rec_type = (unsigned char)strtol(tmp,NULL,16);

            if(rec_type == 1) break;

            chksum = 0;

            if ((rec_addr + rec_length >= UserFlash) && warning) 
            {
               printf("Warning:  HexFile too large for the selected CPU\n");
               warning=0;
            }

            //for(i=0;(i < rec_length) && (i < UserFlash - rec_addr); ++i)
            for(i=0;i < rec_length; ++i)
            {
               memcpy(tmp,&line[(i*2)+9],2); tmp[2] = '\0';
               FlashBuf[rec_addr+i] = (unsigned char)strtol(tmp,NULL,16);
               chksum += FlashBuf[rec_addr+i];
            }
            chksum += rec_length;
            chksum += rec_addr >> 8; chksum += rec_addr & 0xFF;
            chksum += rec_type;
            chksum = (~chksum) + 1;
            
            memcpy(tmp,&line[(i*2)+9],2); tmp[2] = '\0';
            rec_chksum = (unsigned char)strtol(tmp,NULL,16);
            if(rec_chksum != chksum) 
            { 
               printf("Error:  Checksum wrong %s\n",line);   
               //printf("%X %X\n",rec_chksum,chksum);
               fclose(file);
               return -1;
            }
         }
         fclose (file);
         return 0;
      }  
   }    
   return -1 ;
}

// writes the FlashBuf contents into a bianry file until limit is reached
int BinWriter(char *fname)
{
   FILE     *file;
   int      addr;

   if (fname)
   {
      file = fopen(fname,"wb");
      if (!file)
      {
         printf("creating binary file failed!\n");
         return -1 ;
      }
      addr=0;
      while (addr < LimitSize)
      {
         fwrite(&FlashBuf[addr],1,1,file); 
         addr++;  
      }
      fclose(file);
      return 0;
   }
   return -1;
}

// writes the FlashBuf contents into a intel hex file until limit is reached
int HexWriter(char *fname)
{
   FILE     *file;
   char     line[80];
   unsigned char  nbytes,b,chk;
   unsigned short addr;
   int      i;

   if (fname)
   {
      file = fopen(fname,"wt");
      if (!file)
      {
         printf("creating hexfile failed!\n");
         return -1 ;
      }
      addr=0;
      while (addr < LimitSize)
      {
         line[0]= ':';
         line[1]= '1';
         line[2]= '0';         
         chk= 16;
         sprintf(&line[3],"%04X",addr); chk+= addr & 0xFF;chk+= addr >>8 & 0xFF;
         line[7]='0'; line[8]='0';
         for (i=0;i<16;i++)
         {
            sprintf(&line[i*2+9],"%02X",FlashBuf[addr+i]);
            chk+= FlashBuf[addr+i];
         }
         chk = ~chk+1;
         sprintf(&line[i*2+9],"%02X",chk); 
         line[i*2+11]='\0';
         fprintf (file,"%s\n",line);
         addr +=16;
      }
      fprintf (file,":00000001FF\n");
      fclose(file);
      return 0;
   }
   return -1;
}

// TeaEncrypter scramples the complete FlashBuf optionally using msb first
#define TEA_ROUNDS   32L
#define TEA_DELTA    0x9E3779B9L

void TeaEncipher ( const unsigned long *const InData,  // data to encrypt
                         unsigned long *const OutData, // encrypted data
                   const unsigned long *const Key)     // krypt key
{
    unsigned long y   = InData[0];
    unsigned long z   = InData[1];
    unsigned long sum = 0;
    int           n   = TEA_ROUNDS;
    
    if(MsbFirst)
    {  
       y = htonl(InData[0]);
       z = htonl(InData[1]);
    }  
    while( n )
    {
       sum += TEA_DELTA;
       y += (z << 4)+Key[0] ^ z+sum ^ (z >> 5)+Key[1];
       z += (y << 4)+Key[2] ^ y+sum ^ (y >> 5)+Key[3];
       n--;
    }
    if (MsbFirst)
    {
       OutData[0]= htonl(y);
       OutData[1]= htonl(z);
       return;
    }  
    OutData[0]=y;
    OutData[1]=z;
}

void XTeaEncipher( const unsigned long *const InData,  //data to encrypt 
                         unsigned long *const OutData, //encrypted data 
                   const unsigned long *const Key)     //krypt  key
{
    unsigned long y   = InData[0];
    unsigned long z   = InData[1];
    unsigned long sum = 0;
    int           n   = TEA_ROUNDS;
    
    if(MsbFirst)
    {  
       y = htonl(InData[0]);
       z = htonl(InData[1]);
    }  
    
    while( n )
    {
       y   += (((z << 4) ^ (z >> 5)) + z) ^ (sum + Key[sum      & 3]);
       sum += TEA_DELTA;
       z   += (((y << 4) ^ (y >> 5)) + y) ^ (sum + Key[(sum>>11)& 3]);
       n--;
    }
    if (MsbFirst)
    {
       OutData[0]= htonl(y);
       OutData[1]= htonl(z);
       return;
    }  
    OutData[0]=y;
    OutData[1]=z;
}

// main app
int main(int argc, char *argv[])
{
   int  i;
   char option;
   char ch;
   char *p;   
   char fnamearray[MAX_HEX_FILES][MAX_STRING_SIZE];   
   //int  nFiles;
     
   printf("wch auto hex tools \n");
   if (argc==1) {DisplayHelp();exit(0);} 

      
   // collect the options 
   for (i=1;i < argc; i++)
   {
      if ((*argv[i]=='-') || (*argv[i]=='/'))
      {  // all params start with - or /
         argv[i]++;
         option = toupper (*argv[i]++);
         switch (option)
         {
            case 'I'://infile
                 //todo extend to load more than one file
                 strcpy(InFname,argv[i]);                 
                 break; 
            case 'O'://outfile(s)
                 strcpy(OutFname,argv[i]);
                 break;
            case '?':
            case 'H':  
                 DisplayHelp();
                 exit (0); 
                 break;
            case 'K'://enc key
                 strcpy(EncKeyLsb,argv[i]);
                 // convert the msb string
                 EncKeyMsb[3]=EncKeyLsb[0];
                 EncKeyMsb[2]=EncKeyLsb[1];
                 EncKeyMsb[1]=EncKeyLsb[2];
                 EncKeyMsb[0]=EncKeyLsb[3];  

                 EncKeyMsb[7]=EncKeyLsb[4];
                 EncKeyMsb[6]=EncKeyLsb[5];
                 EncKeyMsb[5]=EncKeyLsb[6];
                 EncKeyMsb[4]=EncKeyLsb[7];  
                 
                 EncKeyMsb[11]=EncKeyLsb[8];
                 EncKeyMsb[10]=EncKeyLsb[9];
                 EncKeyMsb[9]=EncKeyLsb[10];
                 EncKeyMsb[8]=EncKeyLsb[11];  

                 EncKeyMsb[15]=EncKeyLsb[12];
                 EncKeyMsb[14]=EncKeyLsb[13];
                 EncKeyMsb[13]=EncKeyLsb[14];
                 EncKeyMsb[12]=EncKeyLsb[15];  
                 break; 
            case 'E'://encrytion on off
                 switch (*argv[i])
                 {
                    case '+':
                         Encode = 1;
                         break;
                    case '-':
                         Encode = 0;
                         break;
                    default:
                         printf("unknown option -e%c%c\n",option,*argv[i]);
                         exit(-1);
                 }         
                 break; 
            case 'F':     // fill with rand val
                 switch (*argv[i])
                 {
                    case '+':
                         FillRandom = 1;
                         break;
                    case '-':
                         FillRandom = 0;
                         break;
                    default:
                         printf("unknown option -f%c\n",argv[i]);
                         exit(-1);
                 }         
                 break; 
            case 'A':  // automatic flashing
                 switch (*argv[i])
                 {
                    case '+':
                         AutoFlash = 1;
                         break;
                    case '-':
                         AutoFlash = 0;
                         break;
                    default:
                         printf("unknown option -a%c%c\n",option,*argv[i]);
                         exit(-1);
                 }         
                 break;
            case 'L'://limit output size
                 LimitSize = strtol (argv[i],NULL,0);
                 break;
            case 'C'://cpu type  
                 WchCpu = strtol (argv[i],NULL,0);
                 switch(WchCpu)
                 {
                    case 0x51:
                         UserFlash = LimitSize = 0x2800;
                         break;  
                    case 0x52:
                    case 0x54:
                         UserFlash = LimitSize = 0x3800;
                         break;                        
                    case 0x58:
                         UserFlash = LimitSize = 0x8000;
                         break;                        
                    case 0x59:
                         UserFlash = LimitSize = 0xF400;                      
                         break;
                    default:
                         printf("unknown cpu -c%d\n",WchCpu);
                         exit (-1);           
                 }
                 break;
            case 'M'://limit output size
                 switch (*argv[i])
                 {
                    case '+':
                         MsbFirst = 1;
                         break;
                    case '-':
                         MsbFirst = 0;
                         break;
                    default:
                         printf("unknown option -k%c%c\n",option,*argv[i]);
                         exit(-1);
                 }         
                 break;
            case 'S': //strong encryption
                 switch (*argv[i])
                 {
                    case '+':
                         //Encode  = 1;//switch on encryption                          
                         //XEncode = 0;
                         StrongEnc = 1;
                         break;
                    case '-':
                         //Encode = 0;
                         StrongEnc = 0;
                         break;
                    default:
                         printf("unknown option -s%c%c\n",option,*argv[i]);
                         exit(-1);
                 }         
                 break;                 
            case 'X': //strong encryption
                 switch (*argv[i])
                 {
                    case '+':
                         XEncode = 1;//switch on encryption                          
                         Encode  = 0;
                         break;
                    case '-':
                         XEncode = 0;
                         break;
                    default:
                         printf("unknown option -x%c%c\n",option,*argv[i]);
                         exit(-1);
                 }         
                 break;                 
            default: 
                 printf("unknown option -%c\n",option);
                 exit(-1);                                  
         }
      }
      else
      {  
         printf ("one or more params wrong!\n");
         DisplayHelp();  
         exit (0);
      }      
   }
   
   //
   // init the buffer with the default 0xFF or random values 
   //
   if(FillRandom)
   {
      randomize();
      for (i=0;i < sizeof(FlashBuf);i++) FlashBuf[i] = random(256);
      printf("  ... buffer filled with random data\n");
   }
   else
   {  
      memset(FlashBuf,0xFF,sizeof(FlashBuf));  
      printf("  ... buffer filled with 0xFF\n");
   }
   // 
   // read the intel hex files into the buffer
   //
   
   // 1. seperate the file names
   memset (fnamearray,'\0',sizeof(fnamearray));
   for (i=0;i<MAX_HEX_FILES;i++)
   {
      p=strrchr(InFname,',');
      if (p) 
      { 
         strcpy(&fnamearray[i][0],p+1); 
         *p = '\0';
      }
      else // the last and possibly only one
      { 
         strcpy(&fnamearray[i][0],InFname); 
         break;
      }
   }
   if(i==MAX_HEX_FILES) {printf("too many hexfiles");exit(-1);}
   // 2. add extensionif its missing 
   for (i=0;i < MAX_HEX_FILES;i++)
   {
      if ((fnamearray[i][0]!='\0') && (strchr(&fnamearray[i][0],'.') == NULL))
        strcat(&fnamearray[i][0],".hex");
   }       
   // 3. read the files into the buffer
   for (i=0;i < MAX_HEX_FILES;i++)
   {
      if (fnamearray[i][0]!='\0') 
      {
         printf("  ... reading hex file %s\n",&fnamearray[i][0]);
         if (HexReader(&fnamearray[i][0]))
         {
             printf ("error in file %s\n",&fnamearray[i][0]);
             exit(-1);
         }
      }
   }
   
   //
   // scramble the contents
   // 

   if (Encode)
   {
      if (MsbFirst)
      {  
         if (StrongEnc)
         { 
            printf("  ... encrypting flashbuf msb first using TEA strong encryption\n");
            for (i=0;i < sizeof(FlashBuf);i++)
            { 
               TeaEncipher((unsigned long*) &FlashBuf[i],
                           (unsigned long*) &FlashBuf[i],
                           (unsigned long*) &EncKeyMsb[0]);
            }
         }
         else
         {
            printf("  ... encrypting flashbuf msb first using TEA fast encryption\n");
            for (i=0;i < sizeof(FlashBuf);i+=8)
            { 
               TeaEncipher((unsigned long*) &FlashBuf[i],
                           (unsigned long*) &FlashBuf[i],
                           (unsigned long*) &EncKeyMsb[0]);
            }
         }    
      }   
      else
      { 
          
         if (StrongEnc)
         { 
            printf("  ... encrypting flashbuf lsb first using TEA strong encryption\n");
            for (i=0;i < sizeof(FlashBuf);i++)
            {
               TeaEncipher((unsigned long*) &FlashBuf[i],
                           (unsigned long*) &FlashBuf[i],
                           (unsigned long*) &EncKeyLsb[0]);
            }
         }   
         else
         {
            printf("  ... encrypting flashbuf lsb first using TEA fast encryption\n");
            for (i=0;i < sizeof(FlashBuf);i+=8)
            {
               TeaEncipher((unsigned long*) &FlashBuf[i],
                           (unsigned long*) &FlashBuf[i],
                           (unsigned long*) &EncKeyLsb[0]);
            }
         }     
      }
   } 

   if (XEncode)
   {
      if (MsbFirst)
      {  
         if (StrongEnc)
         { 
            printf("  ... encrypting flashbuf msb first using XTEA strong encryption\n");
            for (i=0;i < sizeof(FlashBuf);i++)
            { 
               XTeaEncipher((unsigned long*) &FlashBuf[i],
                            (unsigned long*) &FlashBuf[i],
                            (unsigned long*) &EncKeyMsb[0]);
            }
         }
         else
         {
            printf("  ... encrypting flashbuf msb first using XTEA fast encryption\n");
            for (i=0;i < sizeof(FlashBuf);i+=8)
            { 
               XTeaEncipher((unsigned long*) &FlashBuf[i],
                            (unsigned long*) &FlashBuf[i],
                            (unsigned long*) &EncKeyMsb[0]);
            }
         }    
      }   
      else
      { 
          
         if (StrongEnc)
         { 
            printf("  ... encrypting flashbuf lsb first using XTEA strong encryption\n");
            for (i=0;i < sizeof(FlashBuf);i++)
            {
               XTeaEncipher((unsigned long*) &FlashBuf[i],
                            (unsigned long*) &FlashBuf[i],
                            (unsigned long*) &EncKeyLsb[0]);
            }
         }   
         else
         {
            printf("  ... encrypting flashbuf lsb first using XTEA fast encryption\n");
            for (i=0;i < sizeof(FlashBuf);i+=8)
            {
               XTeaEncipher((unsigned long*) &FlashBuf[i],
                            (unsigned long*) &FlashBuf[i],
                            (unsigned long*) &EncKeyLsb[0]);
            }
         }     
      }
   }    
   //
   // create the binary and hex output
   //
   printf("  ... creating output files limit set to 0x%X\n",LimitSize);
   p = strchr(OutFname,'.');
   if (p) *p = '\0';
   BinWriter(strcat (OutFname,".BIN"));

   p = strchr(OutFname,'.');
   if (p) *p = '\0';
   HexWriter(strcat (OutFname,".HEX"));   
   //
   // flash the contents
   //
   if(AutoFlash)
   {
      printf("Warning:  Autoflash option ignored at this time\n");
   } 
   printf("done\n");
   return 0;
}